# Copyright 2016 Jochen Fabricius
# Distribution and modification is encouraged

package provide complex 1.0

namespace eval ::complex {
}

proc ::complex::cset {Z {re 0} {im 0.0}} {
  upvar $Z _Z
  if {$im eq {}} {set im 0}
  set _Z [list [expr {double($re)}] [expr {double($im)}]]
}

proc ::complex::toStr {Z} {
  set op +
  if {[lindex $Z 1] < 0.0} {
    set op -
    set Z [::complex::cconj $Z]
  }
  return [join [list [lindex $Z 0] $op i [lindex $Z 1]] {}]
}

proc ::complex::cadd {A B} {
  set re [expr {[lindex $A 0]+[lindex $B 0]}]
  set im [expr {[lindex $A 1]+[lindex $B 1]}]
  return [list $re $im]
}

proc ::complex::cvadd {A B} {
  if {[set n [llength $A]] != [llength $B]} {
	  return -code error {size mismatch}
  }
  set Z [list]
  foreach X $A Y $B {
	  lappend Z [list [expr {[lindex $X 0]+[lindex $Y 0]}] [expr {[lindex $X 1]+[lindex $Y 1]}]]
  }
  return $Z
}

proc ::complex::cvsum {A} {
	set R 0.0 
	set I 0.0
	foreach {Z} $A {
		lassign $Z re im
		set R [expr {$R+$re}]
		set I [expr {$I+$im}]
	}
	return [list $R $I]
}

proc ::complex::csub {A B} {
  set re [expr {[lindex $A 0]-[lindex $B 0]}]
  set im [expr {[lindex $A 1]-[lindex $B 1]}]
  return [list $re $im]
}

proc ::complex::cvsub {A B} {
  if {[set n [llength $A]] != [llength $B]} {
	  return -code error {size mismatch}
  }
  set Z [list]
  foreach X $A Y $B {
	  lappend Z [list [expr {[lindex $X 0]-[lindex $Y 0]}] [expr {[lindex $X 1]-[lindex $Y 1]}]]
  }
  return $Z
}

proc ::complex::cmul {A B} {
  set re [expr {[lindex $A 0]*[lindex $B 0]-[lindex $A 1]*[lindex $B 1]}]
  set im [expr {[lindex $A 0]*[lindex $B 1]+[lindex $A 1]*[lindex $B 0]}]
  return  [list $re $im]
}

proc ::complex::cvmul {A B} {
  if {[llength $B] == 2 && [llength [lindex $B 0]] == 1} {
	  # b is a scalar
	  lassign $B rb ib
	  set Z [list]
	  foreach X $A {
		  set re [expr {[lindex $X 0]*$rb-[lindex $X 1]*$ib}]
		  set im [expr {[lindex $X 0]*$ib+[lindex $X 1]*$rb}]
		  lappend Z [list $re $im] 
	  }
  } else {
	  if {[llength $A] != [llength $B]} {
		  return -code error {size mismatch}
	  }
	  set Z [list]
	  foreach X $A Y $B {
		  set re [expr {[lindex $X 0]*[lindex $Y 0]-[lindex $X 1]*[lindex $Y 1]}]
		  set im [expr {[lindex $X 0]*[lindex $Y 1]+[lindex $X 1]*[lindex $Y 0]}]
		  lappend Z [list $re $im] 
	  }
  }
  return $Z
}
proc ::complex::cdiv {A B} {
  set det [expr {double([lindex $B 0]**2+[lindex $B 1]**2)}]
  set re [expr {([lindex $A 0]*[lindex $B 0]+[lindex $A 1]*[lindex $B 1])/$det}]
  set im [expr {([lindex $A 0]*[lindex $B 1]-[lindex $A 1]*[lindex $B 0])/$det}]
  return  [list $re $im]
}

proc ::complex::cvdiv {A B} {
  if {[set n [llength $A]] != [llength $B]} {
	  return -code error {size mismatch}
  }
  set Z [list]
  foreach X $A Y $B {
	  set det [expr {double([lindex $X 0]**2+[lindex $Y 1]**2)}]
	  set re [expr {([lindex $X 0]*[lindex $Y 0]+[lindex $X 1]*[lindex $Y 1])/$det}]
	  set im [expr {([lindex $X 0]*[lindex $Y 1]-[lindex $X 1]*[lindex $Y 0])/$det}]
	  lappend Z [list $re $im] 
  }
  return $Z
}

proc ::complex::cabs {Z} {
  return [expr {sqrt([lindex $Z 0]**2+[lindex $Z 1]**2)}]
}

proc ::complex::carg {Z} {
  return [expr {atan2([lindex $Z 1], [lindex $Z 0])}]
}

proc ::complex::cre {Z} {
  return [lindex $Z 0]
}

proc ::complex::cim {Z} {
  return [lindex $Z 1]
}

proc ::complex::cconj {Z} {
  return [list [lindex $Z 0] [expr {-[lindex $Z 1]}]]
}  

proc test {} {
  ::complex::cset A 1 0
  ::complex::cset B 1 1
  puts [::complex::toStr [::complex::cadd $A $B]]
  puts [::complex::toStr [::complex::csub $A $B]]
  puts [::complex::toStr [::complex::cmul $A $B]]
  puts [::complex::toStr [::complex::cdiv $A $B]]
  puts [::complex::cabs $B]
  puts [::complex::carg $B]
}

#test
